"""
This module facilitates access to pixelman frames.
"""
from jarray import array, zeros
from java.lang import String
from cz.ieap.pixelman.shared import SharedFuncsNative
from cz.ieap.jplugininterface.structs import DataTypes, ByteBufferRef
from cz.ieap.pixelman.jmpxloader import Main
from cz.ieap.pixelman.shared.buffersutils import Parameters, BOOL, DOUBLE, I16, U16, I32, U32, STRING, BYTE, CHAR, FLOAT, UCHAR
from java.nio import ByteBuffer, ByteOrder
import sys
import time
import pixelman_regfuncs
import mpxutils

pixelman = Main.mMainClass.getPixelmanFuncs()
SIZE_OF_DOUBLE=8
SIZE_OF_I16=2
SIZE_OF_U32=4


def createFrame(frameName, frameType=DataTypes.TYPE_DOUBLE, width=256, height=256, creator="jscripting"):
    """creates new frame and returns frame object. Frame type, width, height and creator can be specified"""
    frameID = zeros(1, 'i')
    rc = pixelman.createFrame(frameID, width, height, pixelman.FCREATE_ZERO | frameType, "jscripting")
    if rc != 0:
        raise Exception("createFrame(): Cannot create frame: "+str(rc))

    rc = pixelman.setFrameName(frameID[0], frameName)
    if rc != 0:
        raise Exception("createFrame(): Cannot set frame name: "+str(rc))

    frame = MpxFrame()
    frame.loadFromID(frameID[0])
    return frame

def loadFromID(frameID, lockData=False):
    """Loads frame specified by frameID and creates frame object for it that is returned. 
    If lockData is True, frame is locked and it's data buffer is made accessible
    Note: To release frame back to Pixelman unlock() method must be called !!!
    """
    frame = MpxFrame()
    frame.loadFromID(frameID, lockData)
    return frame

def loadFromFile(frameFile, index,  lockData=False):
    """Loads frame from file and returned created object for it. Index specifies index of frame
    in multiframe file. If lockData is True, frame is locked and it's data buffer is made accessible
    Note: To release frame back to Pixelman unlock() method must be called !!!
    """
    frameID = zeros(1, 'i')
    rc = pixelman.loadFrame(frameID, frameFile, index)
    if rc != 0:
        raise Exception("loadFromFile(): Could not load frame "+str(rc))
    frame = MpxFrame()
    frame.loadFromID(frameID[0], lockData)
    return frame

def getFrameCount(fileName):
    """Returns number of frames in specified multiframe file."""
    count = zeros(1,'i')
    if fileName is None:
        raise Exception("MpxFrame.loadFromFile(): fileName cannot be None")
    rc = pixelman.getFrameCount(fileName, count)
    if rc != 0:
        raise Exception("MpxFrame.getFrameCountInFile(): Could not get count: "+str(rc))
    return count[0]

def getAcqFrame(devID, acqIndex):
    """Gets frame from finished acquisition series for specified device. Frame is pecified by 
    acquisition index"""
    frameID = zeros(1, 'i')
    rc = pixelman.mpxCtrlGetFrameID(devID, acqIndex, frameID)
    if rc != 0: 
        raise Exception("MpxFrame.getAcqFrame: Could not get frame ID: "+str(rc))
    return loadFromID(frameID[0])

def getLastAcqFrame(devID):
    """Returns last frame in acquisition series for the device"""
    frameID = mpxutils.getLastAcqFrameID(devID)
    return loadFromID(frameID)

def getStartTime(frameID):
    """Returns start time of the frame specified by frameID. Time is in unix time stamp format (milliseconds from 1.1.1970)"""
    size = array([8], 'l')
    datatype = zeros(1, 'i')
    data = ByteBuffer.allocateDirect(8)
    data.order(ByteOrder.nativeOrder())
    rc = pixelman.getFrameAttrib(frameID, pixelman.FAN_STARTTIME, data, size, datatype);
    if rc != 0:
        sys.exit("Could not get start time: "+str(rc))
    return data.asDoubleBuffer().get(0)

def getAcqTime(frameID):
    """Returns acquisition time of specified frameID"""
    size = array([8], 'l')
    datatype = zeros(1, 'i')
    data = ByteBuffer.allocateDirect(8)
    data.order(ByteOrder.nativeOrder())
    rc = pixelman.getFrameAttrib(frameID, pixelman.FAN_ACQTIME, data, size, datatype);
    if rc != 0:
        sys.exit("Could not get acquisition time: "+str(rc))
    return data.asDoubleBuffer().get(0)


class MpxFrame():
    """Class that facilitates access to frame data and attributes"""
    frameID = 0
    frameType = 0
    size = 0
    width = 0
    height = 0
    locked = False
    frameBuffer = None
 
    def __init__(self):
        pass
    
    def loadFromID(self, frameID, lockData=False):
        width = zeros(1,'l')
        height = zeros(1, 'l')
        frameType = zeros(1, 'i')
        rc = pixelman.getFrameSize(frameID, width, height)
        if rc != 0:
            raise Exception("MpxFrame.getLockedBuffer(): Getting frame size failed: "+str(rc))

        rc = pixelman.getFrameType(frameID, frameType)
        if rc != 0:
            raise Exception("MpxFrame.getLockedBuffer(): Getting frame type failed: "+str(rc))

        self.width = width[0]
        self.height = height[0]
        self.frameType = frameType[0]
        self.frameID = frameID
        if lockData:
            self.lock()
        return

    def loadFromFile(self, fileName, index=0, lockData=False):
        frameID = zeros(1,'i')
        if fileName is None:
            raise Exception("MpxFrame.loadFromFile(): fileName cannot be None")
        rc = pixelman.loadFrame(frameID, fileName, index)
        if rc != 0:
            raise Exception("MpxFrame.loadFromFile(): Could not load frame: "+str(rc))
        self.getFrameByID(frameID[0], lockData)
        return


    def lock(self):
        """Lock frame and gets access to its data. Must be called before get() or put() is called"""
        buff = ByteBufferRef()
        byteSize = zeros(1, 'l')
        rc = pixelman.getLockedFrameBuff(self.frameID, buff, byteSize, 2000)
        if rc != 0:
            raise Exception("MpxFrame.lock(): Getting locked frame buff failed: "+str(rc))
        buff.buffer.order(ByteOrder.nativeOrder())
        if self.frameType == mpxutils.TYPE_DOUBLE:
            self.frameBuffer = buff.buffer.asDoubleBuffer()
            self.size = byteSize[0]/SIZE_OF_DOUBLE

        if self.frameType == mpxutils.TYPE_I16:
            self.frameBuffer = buff.buffer.asShortBuffer()
            self.size = byteSize[0]/SIZE_OF_I16

        if self.frameType == mpxutils.TYPE_U32:
            self.frameBuffer = buff.buffer.asIntBuffer()
            self.size = byteSize[0]/SIZE_OF_U32
        self.locked = True
        return self.frameBuffer

    def unlock(self):
        """Unlocks locked frame and releases it's back to Pixelman. get() and put() cannot be used after the frame is unlocked"""
        rc = pixelman.unlockFrame(self.frameID)
        self.frameBuffer = None
        self.locked = False
        return rc

    def get(self, index):
        """gets value of pixel specified by index. To use this function, frame has to be locked beforehand (lock() function)"""
        if not self.locked: 
            raise Exception("MpxFrame.get(): Frame is not locked, you have to lock it first to work with data !")
        if self.frameBuffer is None:
            raise Exception("MpxFrame.get(): No frame loaded !")
        return self.frameBuffer.get(index)

    def put(self, index, value):
        """Puts value into pixels specified by index. To use this function, frame has to be locked beforehand (lock() function)"""
        if not self.locked:
            raise Exception("MpxFrame.get(): Frame is not locked, you have to lock it first to work with data !")
        if self.frameBuffer is None:
            raise Exception("MpxFrame.put(): No frame loaded !")
        return self.frameBuffer.put(index, value)

    def getPixelCount(self):
        """Gets number of pixels in the frame"""
        return self.size

    def getWidth(self):
        return self.width

    def getHeight(self):
        return self.height

    def getType(self):
        """Gets frame type such as mpxutils.TYPE_I16, mpxutils.TYPE_DOUBLE, mpxutils.TYPE_U32"""
        return self.frameType

    def getName(self):
        """Gets frame name"""
        name = zeros(1, String)
        rc = pixelman.getFrameName(self.frameID, name)
        if rc != 0:
            raise Exception("MpxFrame.getName(): Getting frame name failed: "+str(rc))
        return name[0]

    def getBuffer(self):
        """Gets direct access to buffer variable"""
        return self.frameBuffer

    def close(self):
        """Closes frame, if locked frame is unlocked and it's released from memory if it was loaded from file"""
        if self.frameBuffer is not None:
            self.unlock()

        rc = pixelman.closeFrame(self.frameID, None)
        if rc != 0:
            raise Exception("MpxFrame.deleteFrame(): Could not delete frame: "+str(rc))
        del self.frameBuffer
        self.frameID = 0
        self.size = 0
        self.width = 0
        self.height = 0 

    def getDataCopy(self): 
        """Returns buffer with copy of frame data"""
        width = zeros(1,'l')
        height = zeros(1, 'l')
        rc = pixelman.getFrameSize(self.frameID, width, height)
        if rc != 0:
            raise Exception("MpxFrame.getFrameSize(): Getting frame size failed: "+str(rc))

        frame = ByteBuffer.allocateDirect(width[0]*height[0]*SIZE_OF_DOUBLE)
        frame.order(ByteOrder.nativeOrder())
        doubleBuffer = frame.asDoubleBuffer()
        size = array([width[0]*height[0]*SIZE_OF_DOUBLE], 'l')

        rc = pixelman.getFrameData(self.frameID, frame, size, DataTypes.TYPE_DOUBLE)
        if rc != 0: 
            raise Exception("MpxFrame.getFrameDataCopy(): Could not get frame data: "+str(rc))

        return doubleBuffer

    def addAttrib(self, name, description, type, data):
        """Add new attribute to frame. 
        type - type of the attribute (e.g. mpxutils.TYPE_DOUBLE)
        data - data of the frame 
        """
        buff, length = self.getAttrBuffer(type, data)
        array = zeros(buff.capacity(),'b')
        buff.get(array, 0, len(array));
        pixelman.addFrameAttrib(self.frameID, name, description, type, array, length) 
        
    def setAttrib(self, name, type, data):
        """Sets frame attribute
        type - type of the attribute (e.g. mpxutils.TYPE_DOUBLE)
        """
        buff, length = self.getAttrBuffer(type, data)
        rc = pixelman.setFrameAttrib(self.frameID, name, buff, buff.capacity())
        if rc != 0:
            raise Exception("MpxFrame.setFrameAttrib: Could not set frame attributte \""+name+"\" type: rc="+str(rc))

    def getAttrib(self, name):
        """Gets value of  frame attribute"""
        type = zeros(1,'i')
        size = zeros(1, 'l')
        size[0] = 1024 
        buff = ByteBuffer.allocateDirect(1024)
        buff.order(ByteOrder.nativeOrder())
        rc = pixelman.getFrameAttrib(self.frameID, name, buff, size, type)
        if rc != 0 and rc != -4: #MPXERR_BUFFER_SMALL
            raise Exception("MpxFrame.getFrameAttrib: Could not get frame attributte \""+name+"\" type: rc="+str(rc))

        if rc == -4: #MPXERR_BUFFER_SMALL
            buff = ByteBuffer.allocateDirect(size[0])
            buff.order(ByteOrder.nativeOrder())
            rc = pixelman.getFrameAttrib(self.frameID, name, buff, size, type)

        count = size[0]/SharedFuncsNative.getDataTypeSize(type[0])
        params = None
        if   type[0] == mpxutils.TYPE_BOOL:   params = [BOOL(zeros(count,'z'))]
        elif type[0] == mpxutils.TYPE_DOUBLE: params = [DOUBLE(zeros(count,'d'))]
        elif type[0] == mpxutils.TYPE_FLOAT:  params = [FLOAT(zeros(count,'f'))]
        elif type[0] == mpxutils.TYPE_I16:    params = [I16(zeros(count,'h'))]
        elif type[0] == mpxutils.TYPE_I32:    params = [I32(zeros(count,'i'))]
        elif type[0] == mpxutils.TYPE_I64:    params = [I64(zeros(count,'l'))]
        elif type[0] == mpxutils.TYPE_U16:    params = [U16(zeros(count,'i'))]
        elif type[0] == mpxutils.TYPE_U32:    params = [U32(zeros(count,'l'))]
        elif type[0] == mpxutils.TYPE_U64:    params = [U64(zeros(count,'l'))]
        elif type[0] == mpxutils.TYPE_BYTE:   params = [BYTE(zeros(count,'c'))]
        elif type[0] == mpxutils.TYPE_CHAR:   params = [CHAR(count)]
        elif type[0] == mpxutils.TYPE_UCHAR:  params = [UCHAR(count)] 
        elif type[0] == mpxutils.TYPE_STRING: params = [STRING(count)]
        else:           
            raise Exception("MpxFrame.getFrameAttrib: Unknown frame attribute type: "+str(type[0]))
        
        params = Parameters(params)
        rc = pixelman.getFrameAttrib(self.frameID, name, params.getBuffer(), size, type)
        if rc != 0:
            raise Exception("MpxFrame.getFrameAttrib: Could not get frame attributte \""+name+"\" type: rc="+str(rc))
        
        if count == 1 or type[0] == mpxutils.TYPE_STRING or type[0] == mpxutils.TYPE_CHAR or type[0] == mpxutils.TYPE_UCHAR:
            return params.getParams()[0].getValue()
        else:
            return params.getParams()[0].getValues()

    def getAttrBuffer(self, type, data):
        if not isinstance(data,list): data = [data]
        if type == mpxutils.TYPE_BOOL:   params = map(lambda x: BOOL(x),data)
        if type == mpxutils.TYPE_DOUBLE: params = map(lambda x: DOUBLE(x),data)    
        if type == mpxutils.TYPE_FLOAT:  params = map(lambda x: FLOAT(x),data)    
        if type == mpxutils.TYPE_I16:    params = map(lambda x: I16(x),data)
        if type == mpxutils.TYPE_I32:    params = map(lambda x: I32(x),data)
        if type == mpxutils.TYPE_I64:    params = map(lambda x: I64(x),data)
        if type == mpxutils.TYPE_U16:    params = map(lambda x: U16(x),data)
        if type == mpxutils.TYPE_U32:    params = map(lambda x: U32(x),data)
        if type == mpxutils.TYPE_U64:    params = map(lambda x: U64(x),data)
        if type == mpxutils.TYPE_BYTE:   params = map(lambda x: BYTE(x),data)
        if type == mpxutils.TYPE_CHAR:   params = [CHAR(data[0])]
        if type == mpxutils.TYPE_UCHAR:  params = [UCHAR(data[0])] 
        if type == mpxutils.TYPE_STRING: params = [STRING(data[0])]
        buff = Parameters(params).getBuffer()
        buff.rewind()
        if type == mpxutils.TYPE_STRING or type == mpxutils.TYPE_CHAR or type == mpxutils.TYPE_UCHAR:
            length = len(data[0])
        else:
            length = len(params)
        return buff, length


