# Copyright 2009-2012 IEAP CTU
# Author: Daniel Turecek (daniel.turecek@utef.cvut.cz)
"""
This module facilitate access to Pixelman pixel configuration
"""
from jarray import array, zeros
from java.lang import String
from cz.ieap.pixelman.shared import SharedFuncsNative
from cz.ieap.jplugininterface.structs import DataTypes, ByteBufferRef
from cz.ieap.jplugininterface.structs import HwInfoItem, DevInfo
from cz.ieap.pixelman.jmpxloader import Main
from cz.ieap.pixelman.shared.buffersutils import Parameters, BOOL, DOUBLE, I16, U16, I32, U32, STRING, BYTE, CHAR, FLOAT, UCHAR
from java.nio import ByteBuffer, ByteOrder
import sys
import time
import pixelman_regfuncs
import mpxutils

pixelman = Main.mMainClass.getPixelmanFuncs()

MASKBIT_MASK = 1
TESTBIT_MASK = 2
THL_MASK = 60
LOW_THL_MASK = 28
HIGH_THL_MASK = 224
TPX_MODE_MASK = 192

MODE_MPX = 0
MODE_TOT = 1
MODE_1HIT = 2
MODE_TIMEPIX = 3

def getFromDevice(devID):
    """loads pixel configurtion for all chips from device into pixcfg object and returns it. 
    The matrix has layout as is shown in Preview window."""
    pixcfg = PixCfg() 
    pixcfg.getFromDevice(devID)
    return pixcfg

def loadFromFile(devID, fileName):
    """Loads pixel configuration from a .bpc file and returns pixcfg object."""
    pixcfg = PixCfg() 
    pixcfg.loadFromFile(devID)
    return pixcfg

def resetAllBits(devID):
    """Resets all bits to default values"""
    pixelman.mpxCtrlResetPixelsCfg(devID, -1) # all chips

class PixCfg():
    """Class that facilitates access to pixel configuration data"""
    data = None
    pixCount = 0
    mpxType = 0
    chipCount = 0
    devID = 0

    def __init__(self):
        pass

    def getDevInfo(self, devID):
        """Private function, should not be used."""
        devInfo = DevInfo()
        rc =  pixelman.mpxCtrlGetDevInfo(devID, devInfo)
        if rc != 0: return rc
        self.mpxType = devInfo.mpxType
        self.pixCount = devInfo.pixCount
        self.chipCount = devInfo.numberOfChips
        return 0

    def getFromDevice(self, devID):
        """Private function, should not be used."""
        self.devID = devID
        rc = self.getDevInfo(devID)
        if rc != 0: 
            sys.exit("Could not get device info: "+str(rc))
        size = self.pixCount
        data = ByteBuffer.allocateDirect(size)
        data.order(ByteOrder.nativeOrder())
        rc = pixelman.mpxCtrlGetSuperMatrixPixCfg(devID, data, size)
        if rc != 0:
            sys.exit("Could not get pixel configuration: "+str(rc))
        self.data = data 
        return 0

    def loadFromFile(self, devID, fileName, loadDacs):
        """Private function, should not be used."""
        pixelman.mpxCtrlLoadPixelsCfg(devID, fileName, loadDacs)
        return self.getFromDevice(devID)

    def setToDevice(self, devID):
        """Sets configuration from this object to device"""
        if self.data is None:
            sys.exit("No configuration was loaded -> no configuration to set !")

        rc = pixelman.mpxCtrlSetSuperMatrixPixCfg(devID, self.data, self.data.capacity())
        if rc != 0:
            sys.exit("Could not set pixel configuration: "+str(rc))
        return

    def isMasked(self, pos):
        """Returns True if pixel on position pos is masked"""
        b = self.data.get(pos)
        return ((b & MASKBIT_MASK) == 0)

    def isTestBitOn(self, pos):
        """Returns True if pixel on position pos has test bit on"""
        b = self.data.get(pos)
        return ((b & TESTBIT_MASK) == 0)

    def getThl(self, pos):
        """Returns value of Thl bits for Timepix on position pos (starts from 0)"""
        b = self.data.get(pos)
        return ((b & THL_MASK) >> 2)

    def getLowThl(self, pos):
        """Returns value of LowThl bits for Medipix2 MXR on position pos (starts from 0)"""
        b = self.data.get(pos)
        return ((b & LOW_THL_MASK) >> 2)

    def getHighThl(self, pos):
        """Returns value of HighThl bits for Medipix2 MXR on position pos (starts from 0)"""
        b = self.data.get(pos)
        return ((b & HIGH_THL_MASK) >> 5)

    def getTpxMode(self, pos):
        """Returns value of Timepix mode bits position pos (starts from 0). 
        Mode is one of the values: pixcfg.MODE_MPX, pixcfg.MODE_TOT, pixcfg.MODE_1HIT, pixcfg.MODE_TIMEPIX """
        b = self.data.get(pos)
        return ((b & TPX_MODE_MASK) >> 6)

    def setMasked(self, pos, masked):
        """Sets mask bit on position pos (start from 0). Value == True - masked, Value == False - unmasked"""
        b = int(self.data.get(pos))
        if masked:
            b = (b & ~MASKBIT_MASK)
        else:
            b = b | MASKBIT_MASK
        self.data.put(pos, b)

    def setTestBitOn(self, pos, testBitOn):
        """Sets test bit on position pos (start from 0). Value == True - test bit on, Value == False - test bit off"""
        b = self.data.get(pos)
        if testBitOn:
            b = b & ~TESTBIT_MASK
        else:
            b = b | TESTBIT_MASK
        self.data.put(pos, b)

    def setThl(self, pos, value):
        """Sets value of thl bits for Timepix on position pos"""
        b = self.data.get(pos)
        b = (b & ~THL_MASK) | ((value << 2) & THL_MASK)
        self.data.put(pos, b)

    def setLowThl(self, pos, value):
        """Sets value of low thl bits for Medipix2 MXR on position pos"""
        b = self.data.get(pos)
        b = (b & ~LOW_THL_MASK) | ((value << 2) & LOW_THL_MASK)
        self.data.put(pos, b)

    def setHighThl(self, pos, value):
        """Sets value of high thl bits for Medipix2 MXR on position pos"""
        b = self.data.get(pos)
        b = (b & ~HIGH_THL_MASK) | ((value << 5) & HIGH_THL_MASK)
        self.data.put(pos, b)

    def setTpxMode(self, pos, mode):
        """Sets value of Timepix mode bits position pos (starts from 0). 
        Mode is one of the values: pixcfg.MODE_MPX, pixcfg.MODE_TOT, pixcfg.MODE_1HIT, pixcfg.MODE_TIMEPIX """
        b = self.data.get(pos)
        b = (b & ~TPX_MODE_MASK) | ((mode << 6) & TPX_MODE_MASK)
        self.data.put(pos, b)
